@@ -1,5 +1,5 @@
 /* punycode.h - prototypes for internal punycode functions
-   Copyright (C) 2011-2014 Simon Josefsson
+   Copyright (C) 2011-2017 Simon Josefsson
 
    Libidn2 is free software: you can redistribute it and/or modify it
    under the terms of either:
@@ -38,5 +38,4 @@ extern _IDN2_API int
 _idn2_punycode_decode (size_t input_length,
 		       const char input[],
 		       size_t * output_length,
-		       uint32_t output[],
-		       unsigned char case_flags[]);
+		       uint32_t output[], unsigned char case_flags[]);
