@@ -1,5 +1,5 @@
 /* test-lookup.c --- Self tests for IDNA processing
-   Copyright (C) 2011-2014 Simon Josefsson
+   Copyright (C) 2011-2017 Simon Josefsson
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,9 +22,13 @@
 #include <stdarg.h>
 #include <string.h>
 #include <stdint.h>
+#include <ctype.h>
+#include <errno.h>
 
 #include <idn2.h>
 
+#include "unistr.h"		/* u32_to_u8, u8_to_u32 */
+
 struct idna
 {
   const char *in;
@@ -40,7 +44,7 @@ static const struct idna idna[] = {
   {"..", "..", IDN2_OK}, /* XXX should we disallow this? */
 
   /* U+19DA */
-  {"\xe1\xa7\x9a", "xn--pkf", IDN2_OK},
+  {"\xe1\xa7\x9a", "xn--pkf", IDN2_DISALLOWED},
 
   /* UTC's test vectors. */
 #include "IdnaTest.inc"
@@ -779,26 +783,33 @@ static const struct idna idna[] = {
   },
   {"\xe2\x84\xab", "", IDN2_NOT_NFC},
   {"\xe2\x84\xa6", "", IDN2_NOT_NFC},
+  {
+    /* blåbærgrød.no composed */
+    "\x62\x6c\xc3\xa5\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64\x2e\x6e\x6f", "xn--blbrgrd-fxak7p.no", IDN2_OK
+  },
+  {
+    /* blåbærgrød.no partially decomposed */
+    "\x62\x6c\x61\xcc\x8a\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64\x2e\x6e\x6f", "", IDN2_NOT_NFC
+  },
+  {
+    /* blåbærgrød.no partially decomposed */
+    "\x62\x6c\x61\xcc\x8a\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64\x2e\x6e\x6f", "xn--blbrgrd-fxak7p.no", IDN2_OK, IDN2_NFC_INPUT
+  },
+  {
+    /* blåbærgrød.no partially decomposed */
+    "\x62\x6c\x61\xcc\x8a\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64\x2e\x6e\x6f", "xn--blbrgrd-fxak7p.no", IDN2_OK, IDN2_TRANSITIONAL
+  },
+  {
+    /* blåbærgrød.no partially decomposed */
+    "\x62\x6c\x61\xcc\x8a\x62\xc3\xa6\x72\x67\x72\xc3\xb8\x64\x2e\x6e\x6f", "xn--blbrgrd-fxak7p.no", IDN2_OK, IDN2_NONTRANSITIONAL
+  },
 };
 
-int debug = 1;
-int error_count = 0;
-int break_on_error = 1;
+static int debug = 1;
+static int ok = 0, failed = 0;
+static int break_on_error = 0;
 
-void
-fail (const char *format, ...)
-{
-  va_list arg_ptr;
-
-  va_start (arg_ptr, format);
-  vfprintf (stderr, format, arg_ptr);
-  va_end (arg_ptr);
-  error_count++;
-  if (break_on_error)
-    exit (EXIT_FAILURE);
-}
-
-void
+static void
 hexprint (const char *str, size_t len)
 {
   size_t i;
@@ -816,37 +827,303 @@ hexprint (const char *str, size_t len)
   printf ("\n");
 }
 
-int
-main (void)
+static char *_nextField(char **line)
+{
+  char *s = *line, *e;
+
+  if (!*s)
+    return NULL;
+
+  if (!(e = strpbrk(s, ";#"))) {
+    e = *line += strlen(s);
+  } else {
+    *line = e + (*e == ';');
+    *e = 0;
+  }
+
+  // trim leading and trailing whitespace
+  while (isspace(*s)) s++;
+  while (e > s && isspace(e[-1])) *--e = 0;
+
+  return s;
+}
+
+static int _scan_file(const char *fname, int(*scan)(char *))
+{
+  FILE *fp = fopen(fname, "r");
+  char *buf = NULL, *linep;
+  size_t bufsize = 0;
+  ssize_t buflen;
+  int ret = 0;
+
+  if (!fp) {
+    fprintf(stderr, "Failed to open %s (%d)\n", fname, errno);
+    return -1;
+  }
+
+  while ((buflen = getline(&buf, &bufsize, fp)) >= 0) {
+    linep = buf;
+
+    while (isspace(*linep)) linep++; // ignore leading whitespace
+
+    // strip off \r\n
+    while (buflen > 0 && (buf[buflen] == '\n' || buf[buflen] == '\r'))
+      buf[--buflen] = 0;
+
+    if (!*linep || *linep == '#') continue; // skip empty lines and comments
+
+    if ((ret = scan(linep)))
+      break;
+  }
+
+  free(buf);
+  fclose(fp);
+
+  return ret;
+}
+
+static void
+test_homebrewed(void)
 {
   uint8_t *out;
   size_t i;
   int rc;
 
-  puts ("-----------------------------------------------------------"
-	"-------------------------------------");
-  puts ("                                          IDNA2008 Lookup\n");
-  puts ("  #  Result                    ACE output                  "
-	"             Unicode input");
-  puts ("-----------------------------------------------------------"
-	"-------------------------------------");
   for (i = 0; i < sizeof (idna) / sizeof (idna[0]); i++)
     {
       rc = idn2_lookup_u8 (idna[i].in, &out, idna[i].flags);
-      printf ("%3d  %-25s %-40s %s\n", i, idn2_strerror_name (rc),
+      printf ("%3d  %-25s %-40s %s\n", (int) i, idn2_strerror_name (rc),
 	      rc == IDN2_OK ? idna[i].out : "", idna[i].in);
-      if (rc != idna[i].rc && rc == IDN2_ENCODING_ERROR)
-	printf ("utc bug %d\n", i);
-      else if (rc != idna[i].rc && idna[i].rc != -1)
-	fail ("expected rc %d got rc %d\n", idna[i].rc, rc);
-      else if (rc == IDN2_OK && strcmp (out, idna[i].out) != 0)
-	fail ("expected: %s\ngot: %s\n", idna[i].out, out);
+
+      if (rc != idna[i].rc && rc == IDN2_ENCODING_ERROR) {
+	printf("utc bug\n");
+      } else if (rc != idna[i].rc && idna[i].rc != -1) {
+	failed++;
+	printf("expected rc %d got rc %d\n", idna[i].rc, rc);
+      } else if (rc == IDN2_OK && strcmp (out, idna[i].out) != 0) {
+	failed++;
+	printf("expected: %s\ngot: %s\n", idna[i].out, out);
+      } else
+	ok++;
+
       if (rc == IDN2_OK)
-	free (out);
+	idn2_free (out);
+
+      if (failed && break_on_error)
+	exit (EXIT_FAILURE);
     }
 
+  /* special calls to cover edge cases */
+  if ((rc = idn2_lookup_u8 (NULL, NULL, 0)) != IDN2_OK) {
+    failed++;
+    printf("special #1 failed with %d\n", rc);
+  } else
+    ok++;
+
+  out = (void *) 0x123;
+  if ((rc = idn2_lookup_u8 (NULL, &out, 0)) != IDN2_OK) {
+    failed++;
+    printf("special #2 failed with %d\n", rc);
+  } else if (out) {
+    failed++;
+    printf("special #2 failed with out!=NULL\n");
+  } else
+    ok++;
+
+  if ((rc = idn2_lookup_ul (NULL, NULL, 0)) != IDN2_OK) {
+    failed++;
+    printf("special #3 failed with %d\n", rc);
+  } else
+    ok++;
+
+  out = (void *) 0x123;
+  if ((rc = idn2_lookup_ul (NULL, (char **) &out, 0)) != IDN2_OK) {
+    failed++;
+    printf("special #4 failed with %d\n", rc);
+  } else if (out) {
+    failed++;
+    printf("special #2 failed with out!=NULL\n");
+  } else
+    ok++;
+}
+
+// decode embedded UTF-16/32 sequences
+static char *
+_decodeIdnaTest(uint8_t *src_u8)
+{
+  size_t it2 = 0, len;
+  uint32_t *src;
+
+  // convert UTF-8 to UCS-4 (Unicode))
+  if (!(src = u8_to_u32(src_u8, u8_strlen(src_u8) + 1, NULL, &len))) {
+    printf("u8_to_u32(%s) failed (%d)\n", src_u8, errno);
+    return NULL;
+  }
+
+  // replace escaped UTF-16 incl. surrogates
+  for (size_t it = 0; it < len;) {
+    if (src[it] == '\\' && src[it + 1] == 'u') {
+      src[it2] =
+	((src[it + 2] >= 'A' ? src[it + 2] - 'A' + 10 : src[it + 2] - '0') << 12) +
+	((src[it + 3] >= 'A' ? src[it + 3] - 'A' + 10 : src[it + 3] - '0') << 8) +
+	((src[it + 4] >= 'A' ? src[it + 4] - 'A' + 10 : src[it + 4] - '0') << 4) +
+	(src[it + 5] >= 'A' ? src[it + 5] - 'A' + 10 : src[it + 5] - '0');
+      it += 6;
+
+      if (src[it2] >= 0xD800 && src[it2] <= 0xDBFF) {
+	// high surrogate followed by low surrogate
+	if (src[it] == '\\' && src[it + 1] == 'u') {
+	  uint32_t low =
+	    ((src[it + 2] >= 'A' ? src[it + 2] - 'A' + 10 : src[it + 2] - '0') << 12) +
+	    ((src[it + 3] >= 'A' ? src[it + 3] - 'A' + 10 : src[it + 3] - '0') << 8) +
+	    ((src[it + 4] >= 'A' ? src[it + 4] - 'A' + 10 : src[it + 4] - '0') << 4) +
+	    (src[it + 5] >= 'A' ? src[it + 5] - 'A' + 10 : src[it + 5] - '0');
+	  if (low >= 0xDC00 && low <= 0xDFFF)
+	    src[it2] = 0x10000 + (src[it2] - 0xD800) * 0x400 + (low - 0xDC00);
+	  else
+	    printf("Missing low surrogate\n");
+	  it += 6;
+	} else {
+	  it++;
+	  printf("Missing low surrogate\n");
+	}
+      }
+      it2++;
+    } else
+      src[it2++] = src[it++];
+  }
+
+  // convert UTF-32 to UTF-8
+  uint8_t *tmp=src_u8;
+  src_u8 = u32_to_u8(src, it2, NULL, &len);
+  free(src);
+  if (!src_u8) {
+    printf("u32_to_u8(%s) failed (%d)\n", tmp, errno);
+    return NULL;
+  }
+
+  return src_u8;
+}
+
+static void
+_check_toASCII(char *source, char *expected, int transitional, int expected_toASCII_failure)
+{
+  int rc;
+  char *ace = NULL;
+
+  rc = idn2_lookup_u8 (source, (uint8_t **) &ace, transitional ? IDN2_TRANSITIONAL : IDN2_NONTRANSITIONAL);
+
+  // printf("n=%d expected=%s t=%d got=%s, expected_failure=%d\n", n, expected, transitional, ace ? ace : "", expected_toASCII_failure);
+  if (rc && expected_toASCII_failure) {
+    printf("OK\n");
+    ok++;
+  } else if (rc && !transitional && *expected != '[') {
+    failed++;
+    printf("Failed: _check_toASCII(%s) -> %d (expected 0) %p\n", source, rc, ace);
+  } else if (rc == 0 && !transitional && *expected != '[' && strcmp(expected, ace)) {
+    failed++;
+    printf("Failed: _check_toASCII(%s) -> %s (expected %s) %p\n", source, ace, expected, ace);
+  } else {
+    printf("OK\n");
+    ok++;
+  }
+
+  if (rc == IDN2_OK)
+    idn2_free(ace);
+}
+
+static int
+test_IdnaTest(char *linep)
+{
+  char *type, *source, *toUnicode, *toASCII, *NV8, *org_source;
+  int expected_toASCII_failure;
+#if HAVE_LIBUNISTRING
+  extern int _libunistring_version;
+#endif
+
+  type = _nextField(&linep);
+  org_source = source = _nextField(&linep);
+  toUnicode = _nextField(&linep);
+  toASCII = _nextField(&linep);
+  NV8 = _nextField(&linep); // if set, the input should be disallowed for IDNA2008
+
+  // sigh, these Unicode people really mix UTF-8 and UCS-2/4
+  // quick and dirty translation of '\uXXXX' found in IdnaTest.txt including surrogate handling
+  source = _decodeIdnaTest(source);
+  if (!source)
+    return 0; // some Unicode sequences can't be encoded into UTF-8, skip them
+
+  if (!*toUnicode)
+    toUnicode = source;
+  if (!*toASCII)
+    toASCII = toUnicode;
+  expected_toASCII_failure = NV8 && *NV8;
+
+  printf("##########%s#%s#%s#%s#%s#\n", type, org_source, toUnicode, toASCII, NV8);
+
+#if HAVE_LIBUNISTRING
+  /* 3 tests fail with libunicode <= 0.9.3 - just skip them until we have a newer version installed */
+  /* watch out, libunicode changed versioning scheme up from 0.9.4 */
+  /* If !HAVE_LIBUNISTRING, we use internal gnulib code which works. */
+  if (_libunistring_version <= 9) {
+    if (!strcmp(toASCII, "xn--8jb.xn--etb875g")) {
+      free(source);
+      return 0;
+    }
+  }
+#endif
+
+  if (*type == 'B') {
+    _check_toASCII(source, toASCII, 1, expected_toASCII_failure);
+    _check_toASCII(source, toASCII, 0, expected_toASCII_failure);
+  } else if (*type == 'T') {
+    _check_toASCII(source, toASCII, 1, expected_toASCII_failure);
+  } else if (*type == 'N') {
+    _check_toASCII(source, toASCII, 0, expected_toASCII_failure);
+  } else {
+    printf("Failed: Unknown type '%s'\n", type);
+  }
+
+  free(source);
+
+  if (failed && break_on_error)
+    return 1;
+
+  return 0;
+}
+
+static void
+separator(void)
+{
   puts ("-----------------------------------------------------------"
 	"-------------------------------------");
+}
 
-  return error_count;
+int
+main (int argc, const char *argv[])
+{
+  separator();
+  puts ("                                          IDNA2008 Lookup\n");
+  puts ("  #  Result                    ACE output                  "
+	"             Unicode input");
+  separator();
+
+  test_homebrewed();
+
+  separator();
+
+  // test all IDNA cases from Unicode 9.0.0
+  if (_scan_file(argc == 1 ? SRCDIR"/IdnaTest.txt" : argv[1], test_IdnaTest))
+    return EXIT_FAILURE;
+
+  separator();
+
+  if (failed) {
+    printf("Summary: %d out of %d tests failed\n", failed, ok + failed);
+    return EXIT_FAILURE;
+  }
+
+  printf("Summary: All %d tests passed\n", ok + failed);
+
+  return EXIT_SUCCESS;
 }
